import euclid
import ac
import acsys
import math
import heliSettings

class HeliCar(object):
    """This represents both a car with coordinates and a player (with laps etc. for blue flagging)"""
    name = "player name"
    id = 0
    currentWorldPosition = None
    currentVelocityVector = None
    currentTyreHeadingVector = None
    currentSpeed = 0.0
    splineposition = 0.0
    laps = -1
    playerDistance = 0.0

    relativePositionsAngle=0.0
    relativePositionMeters = None
    centerPositionGui = None

    overlapIndicatorL=0
    overlapIndicatorR=0

    maxOpacity = 0.0

    def __init__(self,carId, name):
        self.name = name
        self.id = carId


    def calcPlayer(self):
        speed = ac.getCarState(self.id, acsys.CS.SpeedKMH)
        if speed < 1:
            self.currentSpeed = 0.0
        else:
            self.currentSpeed = speed

        #get the world position and direction
        x, y, z = ac.getCarState(self.id, acsys.CS.WorldPosition)
        self.currentWorldPosition = euclid.Point2(x, z)
        f, u, l = ac.getCarState(self.id, acsys.CS.Velocity)
        self.currentVelocityVector = euclid.Vector2(f, l).normalize()
        self.laps = ac.getCarState(self.id, acsys.CS.LapCount)
        self.splineposition = ac.getCarState(self.id, acsys.CS.NormalizedSplinePosition)

        #the relative Position (to the player itself) is 0,0
        self.relativePositionMeters = euclid.Point2(0,0)
        self.playerDistance = 0

    def calc(self, player):
        #first, detect wether this car is moving at all. 
        #If not, we don't need further information - it shouldn't be drawn at all
        speed = ac.getCarState(self.id, acsys.CS.SpeedKMH)
        if speed < 1:
            self.currentSpeed = 0.0
        else:
            self.currentSpeed = speed

            #get the world position and direction, already relative to the
            #players world position
            x, y, z = ac.getCarState(self.id, acsys.CS.WorldPosition)
            self.currentWorldPosition = euclid.Point2(x, z)
            f, u, l = ac.getCarState(self.id, acsys.CS.Velocity)
            self.currentVelocityVector = euclid.Vector2(f, l).normalize()
            self.laps = ac.getCarState(self.id, acsys.CS.LapCount)
            self.splineposition = ac.getCarState(self.id, acsys.CS.NormalizedSplinePosition)
        
            self.relativePositionMeters = euclid.Point2(x - player.currentWorldPosition.x, z - player.currentWorldPosition.y)
            self.playerDistance = player.currentWorldPosition.distance(euclid.Point2(x,z))

    def calcDrawingInformation(self, playerVectorReversed):
        #the big goal is the determination of the centerPositionGui of this car,
        #as well as the opacity to draw this one

        #the player car is slightly easier. We start at 0,0, add the offset and multiply with zoom
        if self.id == 0:
            self.centerPositionGui = euclid.Point2((0 + heliSettings.xOff) * heliSettings.guiZoomFactor,(0 + heliSettings.xOff) * heliSettings.guiZoomFactor)
            self.maxOpacity = 1.0
        else:
            #the other cars have to be rotated around the origin, then translated and zoomed

            ####
            #Rotation: We need the angle to the reverted playerVector
            angle =  math.atan2(-1, 0) - math.atan2(playerVectorReversed.y, playerVectorReversed.x)
            #better debugging: to DEG
            angleD = angle * 360 / (2 * math.pi)
            #.  Rotation of our point around the player
            angleR = angleD * math.pi / 180
            cosTheta = math.cos(angleR)
            sinTheta = math.sin(angleR)
            x = cosTheta * self.relativePositionMeters.x - sinTheta * self.relativePositionMeters.y
            y = sinTheta * self.relativePositionMeters.x + cosTheta * self.relativePositionMeters.y

            ####
            #Opacity: How far away is the other car - regarding y (we don't want to fade someone out who is on the same height)
            #self.maxOpacity = self.calcAlpha(self.playerDistance) #would be a circle around the player
            self.maxOpacity = self.calcAlpha(y)

            #####
            #Overlaping indicator
            if heliSettings.showIndicators == 1:
                self.calcOverlap(-x,y)

            ####
            #Translation: Now we add the offsets and multiply with the zoom
            x = x * heliSettings.worldzoom * -1
            y = y * heliSettings.worldzoom * -1
            self.centerPositionGui = euclid.Point2( (x+heliSettings.xOff)*heliSettings.guiZoomFactor, (y+heliSettings.yOff)*heliSettings.guiZoomFactor)

    
    def calcAngleToPlayer(self, playerGuiPosition):

        ####
        #Angle to the player
        dx,dy = self.centerPositionGui.x-playerGuiPosition.x, self.centerPositionGui.y-playerGuiPosition.y
        rads = math.atan2(dy, dx)
        self.relativePositionsAngle = math.degrees(rads) + 90.0

        self.maxOpacity *= self.calcAlphaFromAngle()
    

    def calcAlpha(self, distanceInMeters):

        #alpha is determined by some distance to the player (usually y or x/y distance)        
        alpha = 1 - (math.fabs(distanceInMeters)-heliSettings.opacityThreshold)/(heliSettings.distanceThreshold-heliSettings.opacityThreshold)
        
        if alpha < 0:
            alpha = 0
        elif alpha > 1:
            alpha = 1
            
        return alpha

    def calcOverlap(self, relativeX, distanceInMeters):
        #We seek some indicator if there is a) almost to a little bit and b) substantial overlap on each side
        self.overlapIndicatorL = 0
        self.overlapIndicatorR = 0

        #Substantial= less than a 0.8 car lengths
        if math.fabs(distanceInMeters) < heliSettings.carLength*0.8:
            if relativeX > 0:
                self.overlapIndicatorR = 2
            else:
                self.overlapIndicatorL = 2
        #a bit is less than car length + 20% safety
        elif math.fabs(distanceInMeters) < heliSettings.carLength*1.2:
            if relativeX < 0:
                self.overlapIndicatorL = 1
            else:
                self.overlapIndicatorR = 1

    def calcAlphaFromAngle(self):

        #If configured, we'll reduce alpha in front of the player
        if heliSettings.frontFadeOutArc <= 0:
            return 1.0

        #We assume a car in front is 0, and -90 (left side) or 90 (right side)
        alpha = (math.fabs(self.relativePositionsAngle)- heliSettings.frontFadeOutArc/2.0 + heliSettings.frontFadeAngle/2.0) / 10
        if alpha < 0:
            alpha = 0
        elif alpha > 1:
            alpha = 1
        return alpha

       
    def calcAlphaFromY(self):
        #BEWARE, this is plain wrong. relativePositions.y is still world-aligned, let's stick with the distance instead.

        #here we seek the longditudal(does this word exist?) distance to the player.
        #therefore we'll return to the relativePositionMeters.y, because this is already the distance in meters
        y = math.fabs(self.relativePositionMeters.y)
        if y < heliSettings.opacityThreshold:
            #inside the "draw always-circle"?
            return 1.0
        if y > heliSettings.distanceThreshold:
            #outside the "don't bother me"-circle? Shouldn't be possible anyway
            return 0.0

        #else we'll just get a linear decreasing value.
        return 1 - ( y / (heliSettings.distanceThreshold - heliSettings.opacityThreshold - heliSettings.carLength))

    def drawYourself(self, r, g, b):
        try:
            l = heliSettings.carLength / 2 * heliSettings.worldzoom * heliSettings.guiZoomFactor
            w = heliSettings.carWidth / 2 * heliSettings.worldzoom * heliSettings.guiZoomFactor
            ac.glColor4f(r,g,b,self.maxOpacity * heliSettings.maxiumAlpha)
            ac.glBegin(acsys.GL.Quads)
            ac.glVertex2f(self.centerPositionGui.x - w, self.centerPositionGui.y + l)
            ac.glVertex2f(self.centerPositionGui.x + w, self.centerPositionGui.y + l)
            ac.glVertex2f(self.centerPositionGui.x + w, self.centerPositionGui.y - l)
            ac.glVertex2f(self.centerPositionGui.x - w, self.centerPositionGui.y - l)
            ac.glEnd()
        except Exception as e:
            ac.log("helicorsa::drawYourself self.centerPositionGui == None for car {0}, Error={1}".format(self.id, e))

    def drawYourselfRoundedWithBorders(self, r, g, b, br, bg, bb):
        try:
            l = heliSettings.carLength / 2 * heliSettings.worldzoom * heliSettings.guiZoomFactor
            w = heliSettings.carWidth / 2 * heliSettings.worldzoom * heliSettings.guiZoomFactor
            
            #first we draw a background quad in the original measurements and the b-colors
            #it's important we use rounded (or trunc'd) values, or the drawing will wobber around
            ac.glColor4f(br,bg,bb,self.maxOpacity * heliSettings.maxiumAlpha)
            ac.glBegin(acsys.GL.Quads)
            ac.glVertex2f(round(self.centerPositionGui.x - w), round(self.centerPositionGui.y + l))
            ac.glVertex2f(round(self.centerPositionGui.x + w), round(self.centerPositionGui.y + l))
            ac.glVertex2f(round(self.centerPositionGui.x + w), round(self.centerPositionGui.y - l))
            ac.glVertex2f(round(self.centerPositionGui.x - w), round(self.centerPositionGui.y - l))
            ac.glEnd()

            #then the same quad in the target (fill) color with slightly reduced bounds
            t = heliSettings.carBorderThickness
            ac.glColor4f(r,g,b,self.maxOpacity * heliSettings.maxiumAlpha)
            ac.glBegin(acsys.GL.Quads)
            ac.glVertex2f(round(self.centerPositionGui.x - w + t), round(self.centerPositionGui.y + l  - t))
            ac.glVertex2f(round(self.centerPositionGui.x + w - t), round(self.centerPositionGui.y + l - t))
            ac.glVertex2f(round(self.centerPositionGui.x + w - t), round(self.centerPositionGui.y - l + t))
            ac.glColor4f(r,g,b, self.maxOpacity * heliSettings.maxiumAlpha * 0.5) # this will give some fancy graphics effect
            ac.glVertex2f(round(self.centerPositionGui.x - w + t), round(self.centerPositionGui.y - l + t))
            ac.glEnd()

        except Exception as e:
            ac.log("helicorsa::drawYourselfRoundedWithBorders self.centerPositionGui == None for car {0}, Error={1}".format(self.id, e))

